@@ -19,11 +19,11 @@ Struct::Path - Path for nested structures where path is also a structure
 
 =head1 VERSION
 
-Version 0.64
+Version 0.65
 
 =cut
 
-our $VERSION = '0.64';
+our $VERSION = '0.65';
 
 =head1 SYNOPSIS
 
@@ -43,11 +43,11 @@ our $VERSION = '0.64';
 
     @list = slist($s);                              # get all paths and their values
     # @list == (
-    #    [[[0]],0],
-    #    [[[1]],1],
-    #    [[[2],{keys => ['2a']},{keys => ['2aa']}],'2aav'],
-    #    [[[2],{keys => ['2a']},{keys => ['2ab']}],'2abv'],
-    #    [[[3]],undef]
+    #    [[[0]],\0],
+    #    [[[1]],\1],
+    #    [[[2],{keys => ['2a']},{keys => ['2aa']}],\'2aav'],
+    #    [[[2],{keys => ['2a']},{keys => ['2ab']}],\'2abv'],
+    #    [[[3]],\undef]
     # )
 
     @r = spath($s, [ [3,0,1] ]);                    # get refs to values by paths
@@ -115,14 +115,13 @@ Returns true value if step contains filter or specified all keys/items or key re
 =cut
 
 sub is_implicit_step {
-    my $step = shift;
 
-    if (ref $step eq 'ARRAY') {
-        return 1 unless @{$step};
-    } elsif (ref $step eq 'HASH') {
-        return 1 if (exists $step->{regs} and @{$step->{regs}});
-        return 1 unless (exists $step->{keys});
-        return 1 unless (@{$step->{keys}});
+    if (ref $_[0] eq 'ARRAY') {
+        return 1 unless (@{$_[0]});
+    } elsif (ref $_[0] eq 'HASH') {
+        return 1 if (exists $_[0]->{regs} and @{$_[0]->{regs}});
+        return 1 unless (exists $_[0]->{keys});
+        return 1 unless (@{$_[0]->{keys}});
     } else { # coderefs
         return 1;
     }
@@ -132,7 +131,7 @@ sub is_implicit_step {
 
 =head2 slist
 
-Returns list of paths and their values from structure.
+Returns list of paths and references to their values from structure.
 
     @list = slist($struct, %opts)
 
@@ -151,32 +150,37 @@ Don't dive into structure deeper than defined level.
 sub slist($;@) {
     my ($struct, %opts) = @_;
 
-    my @in = [[], $struct]; # init: [path, lastref]
+    my @in = [[], \$struct]; # init: [path, ref]
     return @in if (defined $opts{depth} and $opts{depth} < 1);
 
-    my @out;
+    my (@out, $p, @unres);
 
-    while (my $p = shift @in) {
-        my @unres; # not fully resolved
-
-        if (ref $p->[1] eq 'HASH' and keys %{$p->[1]}) {
-            for my $k (sort keys %{$p->[1]}) {
-                push @unres, [[@{$p->[0]}, {keys => [$k]}], $p->[1]->{$k}];
+    while ($p = shift @in) {
+        if (ref ${$p->[1]} eq 'HASH' and keys %{${$p->[1]}}) {
+            for (sort keys %{${$p->[1]}}) {
+                push @unres, [
+                    [@{$p->[0]}, {keys => [$_]}],   # path
+                    \${$p->[1]}->{$_}               # ref
+                ];
             }
-        } elsif (ref $p->[1] eq 'ARRAY' and @{$p->[1]}) {
-            for (my $i = 0; $i < @{$p->[1]}; $i++) {
-                push @unres, [[@{$p->[0]}, [$i]], $p->[1]->[$i]];
+        } elsif (ref ${$p->[1]} eq 'ARRAY' and @{${$p->[1]}}) {
+            for (0 .. $#{${$p->[1]}}) {
+                push @unres, [
+                    [@{$p->[0]}, [$_]],             # path
+                    \${$p->[1]}->[$_]               # ref
+                ];
             }
         } else {
             push @out, $p;
         }
 
-        if (defined $opts{depth} and @unres and @{$unres[0]->[0]} >= $opts{depth}) {
-            push @out, @unres;
-            next;
+        if (@unres) {
+            if ($opts{depth} and @{$unres[0]->[0]} >= $opts{depth}) {
+                push @out, splice @unres;
+            } else {
+                unshift @in, splice @unres; # iterate deeper
+            }
         }
-
-        unshift @in, @unres; # reiterate
     }
 
     return @out;
@@ -214,11 +218,12 @@ Croak if at least one element, specified in path, absent in the struct.
 
 sub spath($$;@) {
     my ($struct, $path, %opts) = @_;
-    croak "Path must be arrayref" unless (ref $path eq 'ARRAY');
-    my @out = (ref $struct eq 'ARRAY' or ref $struct eq 'HASH' or not ref $struct) ?
-        [[], [\$struct]] : [[], [$struct]]; # init stacks
 
+    croak "Path must be arrayref" unless (ref $path eq 'ARRAY');
+
+    my @out = [[], [(ref $struct eq 'ARRAY' or ref $struct eq 'HASH' or not ref $struct) ? \$struct : $struct]];
     my $sc = 0; # step counter
+
     for my $step (@{$path}) {
         my @new;
         if (ref $step eq 'ARRAY') {
@@ -278,7 +283,7 @@ sub spath($$;@) {
                             }
                         } else {
                             for my $g (@{$step->{regs}}) {
-                                push @keys, grep { $_ =~ $g} keys %{${$r->[1]->[-1]}};
+                                push @keys, grep { $_ =~ $g } keys %{${$r->[1]->[-1]}};
                             }
                         }
                     }
@@ -309,6 +314,7 @@ sub spath($$;@) {
         $_->[1] = $opts{deref} ? ${pop @{$_->[1]}} : pop @{$_->[1]};
         $_ = $_->[1] unless ($opts{paths});
     } @out;
+
     return @out;
 }
 
@@ -334,17 +340,13 @@ sub spath_delta($$) {
         last unless (ref $frst->[$i] eq ref $scnd->[$i]);
         if (ref $frst->[$i] eq 'ARRAY') {
             last unless (@{$frst->[$i]} == @{$scnd->[$i]});
-            my $j = 0;
-            while ($j < @{$frst->[$i]}) {
+            for my $j (0 .. $#{$frst->[$i]}) {
                 last MAIN unless ($frst->[$i]->[$j] == $scnd->[$i]->[$j]);
-                $j++;
             }
         } elsif (ref $frst->[$i] eq 'HASH') {
             last unless (@{$frst->[$i]->{keys}} == @{$scnd->[$i]->{keys}});
-            my $j = 0;
-            while ($j < @{$frst->[$i]->{keys}}) {
+            for my $j (0 .. $#{$frst->[$i]->{keys}}) {
                 last MAIN unless ($frst->[$i]->{keys}->[$j] eq $scnd->[$i]->{keys}->[$j]);
-                $j++;
             }
         } else {
             croak "Unsupported thing in the path (step #$i)";
